# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""WBF for test-time augmentation."""
import tensorflow as tf


def vectorized_iou(clusters, detection):
  """Calculates the ious for box with each element of clusters."""
  x11, y11, x12, y12 = tf.split(clusters[:, 1:5], 4, axis=1)
  x21, y21, x22, y22 = tf.split(detection[1:5], 4)

  xa = tf.maximum(x11, x21)
  ya = tf.maximum(y11, y21)
  xb = tf.minimum(x12, x22)
  yb = tf.minimum(y12, y22)

  inter_area = tf.maximum((xb - xa), 0) * tf.maximum((yb - ya), 0)

  boxa_area = (x12 - x11) * (y12 - y11)
  boxb_area = (x22 - x21) * (y22 - y21)

  iou = inter_area / (boxa_area + boxb_area - inter_area)

  return iou


def find_matching_cluster(clusters, detection):
  """Returns the index of the highest iou matching cluster for detection."""
  if not clusters:
    return -1
  ious = vectorized_iou(tf.stack(clusters), detection)
  ious = tf.reshape(ious, [len(clusters)])
  if tf.math.reduce_max(ious) < 0.55:
    # returns -1 if no iou is higher than 0.55.
    return -1
  return tf.argmax(ious)


def weighted_average(samples, weights):
  return tf.math.reduce_sum(samples * weights) / tf.math.reduce_sum(weights)


def average_detections(detections, num_models):
  """Takes a list of detections and returns the average, both in box co-ordinates and confidence."""
  num_detections = len(detections)
  detections = tf.stack(detections)
  return [
      detections[0][0],
      weighted_average(detections[:, 1], detections[:, 5]),
      weighted_average(detections[:, 2], detections[:, 5]),
      weighted_average(detections[:, 3], detections[:, 5]),
      weighted_average(detections[:, 4], detections[:, 5]),
      tf.math.reduce_mean(detections[:, 5]) * min(1, num_detections/num_models),
      detections[0][6],
  ]


def ensemble_detections(params, detections, num_models):
  """Ensembles a group of detections by clustering the detections and returning the average of the clusters."""
  all_clusters = []

  for cid in range(params['num_classes']):
    indices = tf.where(tf.equal(detections[:, 6], cid))
    if indices.shape[0] == 0:
      continue
    class_detections = tf.gather_nd(detections, indices)

    clusters = []
    cluster_averages = []
    for d in class_detections:
      cluster_index = find_matching_cluster(cluster_averages, d)
      if cluster_index == -1:
        clusters.append([d])
        cluster_averages.append(average_detections([d], num_models))
      else:
        clusters[cluster_index].append(d)
        cluster_averages[cluster_index] = average_detections(
            clusters[cluster_index], num_models)

    all_clusters.extend(cluster_averages)

  all_clusters.sort(reverse=True, key=lambda d: d[5])
  return tf.stack(all_clusters)



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Model specification for object detection."""

import collections
import functools
import os
import tempfile
from typing import Optional, Tuple, Dict

from absl import logging
import tensorflow as tf
from tensorflow_examples.lite.model_maker.core import compat
from tensorflow_examples.lite.model_maker.core.api.api_util import mm_export
from tensorflow_examples.lite.model_maker.core.data_util import object_detector_dataloader
from tensorflow_examples.lite.model_maker.core.task import configs
from tensorflow_examples.lite.model_maker.core.task.model_spec import util

from tensorflow_examples.lite.model_maker.third_party.efficientdet import coco_metric
from tensorflow_examples.lite.model_maker.third_party.efficientdet import hparams_config
from tensorflow_examples.lite.model_maker.third_party.efficientdet import utils
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import efficientdet_keras
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import eval_tflite
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import label_util
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import postprocess
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import train
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import train_lib

# Number of calibration steps for full integer quantization. 500 steps are
# enough to get a reasonable post-quantization result.
_NUM_CALIBRATION_STEPS = 500


def _get_ordered_label_map(
    label_map: Optional[Dict[int, str]]) -> Optional[Dict[int, str]]:
  """Gets label_map as an OrderedDict instance with ids sorted."""
  if not label_map:
    return label_map
  ordered_label_map = collections.OrderedDict()
  for idx in sorted(label_map.keys()):
    ordered_label_map[idx] = label_map[idx]
  return ordered_label_map


class ExportModel(efficientdet_keras.EfficientDetModel):
  """Model to be exported as SavedModel/TFLite format."""

  def __init__(self,
               model: tf.keras.Model,
               config: hparams_config.Config,
               pre_mode: Optional[str] = 'infer',
               post_mode: Optional[str] = 'global',
               name: Optional[str] = ''):
    """Initilizes an instance with the keras model and pre/post_mode paramaters.

    Args:
      model: The EfficientDetNet model used for training which doesn't have pre
        and post processing.
      config: Model configuration.
      pre_mode: Pre-processing Mode, must be {None, 'infer'}.
      post_mode: Post-processing Mode, must be {None, 'global', 'per_class',
        'tflite'}.
      name: Model name.
    """
    super(efficientdet_keras.EfficientDetNet, self).__init__(name=name)
    self.model = model
    self.config = config
    self.pre_mode = pre_mode
    self.post_mode = post_mode

  @tf.function
  def __call__(self, inputs: tf.Tensor):
    """Calls this model.

    Args:
      inputs: a tensor with common shape [batch, height, width, channels].

    Returns:
      the output tensor list.
    """
    config = self.config

    # Preprocess.
    inputs, scales = self._preprocessing(inputs, config.image_size,
                                         config.mean_rgb, config.stddev_rgb,
                                         self.pre_mode)
    # Network.
    outputs = self.model(inputs, training=False)

    # Postprocess for detection.
    det_outputs = self._postprocess(outputs[0], outputs[1], scales,
                                    self.post_mode)
    outputs = det_outputs + outputs[2:]

    return outputs


@mm_export('object_detector.EfficientDetSpec')
class EfficientDetModelSpec(object):
  """A specification of the EfficientDet model."""

  compat_tf_versions = compat.get_compat_tf_versions(2)

  def __init__(self,
               model_name: str,
               uri: str,
               hparams: str = '',
               model_dir: Optional[str] = None,
               epochs: int = 50,
               batch_size: int = 64,
               steps_per_execution: int = 1,
               moving_average_decay: int = 0,
               var_freeze_expr: str = '(efficientnet|fpn_cells|resample_p6)',
               tflite_max_detections: int = 25,
               strategy: Optional[str] = None,
               tpu: Optional[str] = None,
               gcp_project: Optional[str] = None,
               tpu_zone: Optional[str] = None,
               use_xla: bool = False,
               profile: bool = False,
               debug: bool = False,
               tf_random_seed: int = 111111,
               verbose: int = 0) -> None:
    """Initialze an instance with model paramaters.

    Args:
      model_name: Model name.
      uri: TF-Hub path/url to EfficientDet module.
      hparams: Hyperparameters used to overwrite default configuration. Can be
        1) Dict, contains parameter names and values; 2) String, Comma separated
        k=v pairs of hyperparameters; 3) String, yaml filename which's a module
        containing attributes to use as hyperparameters.
      model_dir: The location to save the model checkpoint files.
      epochs: Default training epochs.
      batch_size: Training & Evaluation batch size.
      steps_per_execution: Number of steps per training execution.
      moving_average_decay: Float. The decay to use for maintaining moving
        averages of the trained parameters.
      var_freeze_expr: Expression to freeze variables.
      tflite_max_detections: The max number of output detections in the TFLite
        model.
      strategy:  A string specifying which distribution strategy to use.
        Accepted values are 'tpu', 'gpus', None. tpu' means to use TPUStrategy.
        'gpus' mean to use MirroredStrategy for multi-gpus. If None, use TF
        default with OneDeviceStrategy.
      tpu: The Cloud TPU to use for training. This should be either the name
        used when creating the Cloud TPU, or a grpc://ip.address.of.tpu:8470
          url.
      gcp_project: Project name for the Cloud TPU-enabled project. If not
        specified, we will attempt to automatically detect the GCE project from
        metadata.
      tpu_zone: GCE zone where the Cloud TPU is located in. If not specified, we
        will attempt to automatically detect the GCE project from metadata.
      use_xla: Use XLA even if strategy is not tpu. If strategy is tpu, always
        use XLA, and this flag has no effect.
      profile: Enable profile mode.
      debug: Enable debug mode.
      tf_random_seed: Fixed random seed for deterministic execution across runs
        for debugging.
      verbose: verbosity mode for `tf.keras.callbacks.ModelCheckpoint`, 0 or 1.
    """
    self.model_name = model_name
    self.uri = uri
    self.batch_size = batch_size
    config = hparams_config.get_efficientdet_config(model_name)
    config.override(hparams)
    config.image_size = utils.parse_image_size(config.image_size)
    config.var_freeze_expr = var_freeze_expr
    config.moving_average_decay = moving_average_decay
    config.tflite_max_detections = tflite_max_detections
    if epochs:
      config.num_epochs = epochs

    if use_xla and strategy != 'tpu':
      tf.config.optimizer.set_jit(True)
      for gpu in tf.config.list_physical_devices('GPU'):
        tf.config.experimental.set_memory_growth(gpu, True)

    if debug:
      tf.config.experimental_run_functions_eagerly(True)
      tf.debugging.set_log_device_placement(True)
      os.environ['TF_DETERMINISTIC_OPS'] = '1'
      tf.random.set_seed(tf_random_seed)
      logging.set_verbosity(logging.DEBUG)

    if strategy == 'tpu':
      tpu_cluster_resolver = tf.distribute.cluster_resolver.TPUClusterResolver(
          tpu, zone=tpu_zone, project=gcp_project)
      tf.config.experimental_connect_to_cluster(tpu_cluster_resolver)
      tf.tpu.experimental.initialize_tpu_system(tpu_cluster_resolver)
      ds_strategy = tf.distribute.TPUStrategy(tpu_cluster_resolver)
      logging.info('All devices: %s', tf.config.list_logical_devices('TPU'))
      tf.config.set_soft_device_placement(True)
    elif strategy == 'gpus':
      ds_strategy = tf.distribute.MirroredStrategy()
      logging.info('All devices: %s', tf.config.list_physical_devices('GPU'))
    else:
      if tf.config.list_physical_devices('GPU'):
        ds_strategy = tf.distribute.OneDeviceStrategy('device:GPU:0')
      else:
        ds_strategy = tf.distribute.OneDeviceStrategy('device:CPU:0')

    self.ds_strategy = ds_strategy

    if model_dir is None:
      model_dir = tempfile.mkdtemp()
    params = dict(
        profile=profile,
        model_name=model_name,
        steps_per_execution=steps_per_execution,
        model_dir=model_dir,
        strategy=strategy,
        batch_size=batch_size,
        tf_random_seed=tf_random_seed,
        debug=debug,
        verbose=verbose)
    config.override(params, True)
    self.config = config

    # set mixed precision policy by keras api.
    precision = utils.get_precision(config.strategy, config.mixed_precision)
    tf.keras.mixed_precision.set_global_policy(precision)

  def create_model(self) -> tf.keras.Model:
    """Creates the EfficientDet model."""
    return train_lib.EfficientDetNetTrainHub(
        config=self.config, hub_module_url=self.uri)

  def train(self,
            model: tf.keras.Model,
            train_dataset: tf.data.Dataset,
            steps_per_epoch: int,
            val_dataset: Optional[tf.data.Dataset],
            validation_steps: int,
            epochs: Optional[int] = None,
            batch_size: Optional[int] = None,
            val_json_file: Optional[str] = None) -> tf.keras.Model:
    """Run EfficientDet training."""
    config = self.config
    if not epochs:
      epochs = config.num_epochs

    if not batch_size:
      batch_size = config.batch_size

    config.update(
        dict(
            steps_per_epoch=steps_per_epoch,
            eval_samples=batch_size * validation_steps,
            val_json_file=val_json_file,
            batch_size=batch_size))
    train.setup_model(model, config)
    train.init_experimental(config)
    model.fit(
        train_dataset,
        epochs=epochs,
        steps_per_epoch=steps_per_epoch,
        callbacks=train_lib.get_callbacks(config.as_dict(), val_dataset),
        validation_data=val_dataset,
        validation_steps=validation_steps)
    return model

  def _get_evaluator_and_label_map(
      self, json_file: str
  ) -> Tuple[coco_metric.EvaluationMetric, Optional[Dict[int, str]]]:
    """Gets evaluator and label_map for evaluation."""
    label_map = label_util.get_label_map(self.config.label_map)
    # Sorts label_map.keys since pycocotools.cocoeval uses sorted catIds
    # (category ids) in COCOeval class.
    label_map = _get_ordered_label_map(label_map)

    evaluator = coco_metric.EvaluationMetric(
        filename=json_file, label_map=label_map)

    evaluator.reset_states()
    return evaluator, label_map

  def _get_metric_dict(self, evaluator: coco_metric.EvaluationMetric,
                       label_map: collections.OrderedDict) -> Dict[str, float]:
    """Gets the metric dict for evaluation."""
    metrics = evaluator.result(log_level=tf.compat.v1.logging.INFO)
    metric_dict = {}
    for i, name in enumerate(evaluator.metric_names):
      metric_dict[name] = metrics[i]

    if label_map:
      for i, cid in enumerate(label_map.keys()):
        name = 'AP_/%s' % label_map[cid]
        metric_dict[name] = metrics[i + len(evaluator.metric_names)]
    return metric_dict

  def evaluate(self,
               model: tf.keras.Model,
               dataset: tf.data.Dataset,
               steps: int,
               json_file: Optional[str] = None) -> Dict[str, float]:
    """Evaluate the EfficientDet keras model.

    Args:
      model: The keras model to be evaluated.
      dataset: tf.data.Dataset used for evaluation.
      steps: Number of steps to evaluate the model.
      json_file: JSON with COCO data format containing golden bounding boxes.
        Used for validation. If None, use the ground truth from the dataloader.
        Refer to
        https://towardsdatascience.com/coco-data-format-for-object-detection-a4c5eaf518c5
          for the description of COCO data format.

    Returns:
      A dict contains AP metrics.
    """
    evaluator, label_map = self._get_evaluator_and_label_map(json_file)
    dataset = dataset.take(steps)

    @tf.function
    def _get_detections(images, labels):
      cls_outputs, box_outputs = model(images, training=False)
      detections = postprocess.generate_detections(self.config, cls_outputs,
                                                   box_outputs,
                                                   labels['image_scales'],
                                                   labels['source_ids'])
      tf.numpy_function(evaluator.update_state, [
          labels['groundtruth_data'],
          postprocess.transform_detections(detections)
      ], [])

    dataset = self.ds_strategy.experimental_distribute_dataset(dataset)
    progbar = tf.keras.utils.Progbar(steps)
    for i, (images, labels) in enumerate(dataset):
      self.ds_strategy.run(_get_detections, (images, labels))
      progbar.update(i + 1)
    print()

    metric_dict = self._get_metric_dict(evaluator, label_map)
    return metric_dict

  def evaluate_tflite(self,
                      tflite_filepath: str,
                      dataset: tf.data.Dataset,
                      steps: int,
                      json_file: Optional[str] = None) -> Dict[str, float]:
    """Evaluate the EfficientDet TFLite model.

    Args:
      tflite_filepath: File path to the TFLite model.
      dataset: tf.data.Dataset used for evaluation.
      steps: Number of steps to evaluate the model.
      json_file: JSON with COCO data format containing golden bounding boxes.
        Used for validation. If None, use the ground truth from the dataloader.
        Refer to
        https://towardsdatascience.com/coco-data-format-for-object-detection-a4c5eaf518c5
          for the description of COCO data format.

    Returns:
      A dict contains AP metrics.
    """
    # TODO(b/182441458): Use the task library for evaluation instead once it
    # supports python interface.
    evaluator, label_map = self._get_evaluator_and_label_map(json_file)
    dataset = dataset.take(steps)

    lite_runner = eval_tflite.LiteRunner(tflite_filepath, only_network=False)
    progbar = tf.keras.utils.Progbar(steps)
    for i, (images, labels) in enumerate(dataset):
      # Get the output result after post-processing NMS op.
      _, nms_scores, nms_classes, nms_boxes = lite_runner.run(images)

      # CLASS_OFFSET is used since label_id for `background` is 0 in label_map
      # while it's not actually included the model. We don't need to add the
      # offset in the Android application.
      nms_classes += postprocess.CLASS_OFFSET

      height, width = utils.parse_image_size(self.config.image_size)
      normalize_factor = tf.constant([height, width, height, width],
                                     dtype=tf.float32)
      nms_boxes *= normalize_factor
      if labels['image_scales'] is not None:
        scales = tf.expand_dims(tf.expand_dims(labels['image_scales'], -1), -1)
        nms_boxes = nms_boxes * tf.cast(scales, nms_boxes.dtype)
      detections = postprocess.generate_detections_from_nms_output(
          nms_boxes, nms_classes, nms_scores, labels['source_ids'])

      detections = postprocess.transform_detections(detections)
      evaluator.update_state(labels['groundtruth_data'].numpy(),
                             detections.numpy())
      progbar.update(i + 1)
    print()

    metric_dict = self._get_metric_dict(evaluator, label_map)
    return metric_dict

  def export_saved_model(self,
                         model: tf.keras.Model,
                         saved_model_dir: str,
                         batch_size: Optional[int] = None,
                         pre_mode: Optional[str] = 'infer',
                         post_mode: Optional[str] = 'global') -> None:
    """Saves the model to Tensorflow SavedModel.

    Args:
      model: The EfficientDetNet model used for training which doesn't have pre
        and post processing.
      saved_model_dir: Folder path for saved model.
      batch_size: Batch size to be saved in saved_model.
      pre_mode: Pre-processing Mode in ExportModel, must be {None, 'infer'}.
      post_mode: Post-processing Mode in ExportModel, must be {None, 'global',
        'per_class', 'tflite'}.
    """
    config = self.config
    # Sets the keras model optimizer to None when exporting to saved model.
    # Otherwise, it fails with `NotImplementedError`: "Learning rate schedule
    # must override get_config".
    original_optimizer = model.optimizer
    model.optimizer = None
    # Creates ExportModel which has pre and post processing.
    export_model = ExportModel(model, config, pre_mode, post_mode)

    # Gets tf.TensorSpec.
    if pre_mode is None:
      # Input is the preprocessed image that's already resized to a certain
      # input shape.
      input_spec = tf.TensorSpec(
          shape=[batch_size, *config.image_size, 3],
          dtype=tf.float32,
          name='images')
    else:
      # Input is that raw image that can be in any input shape,
      input_spec = tf.TensorSpec(
          shape=[batch_size, None, None, 3], dtype=tf.uint8, name='images')

    tf.saved_model.save(
        export_model,
        saved_model_dir,
        signatures=export_model.__call__.get_concrete_function(input_spec))
    model.optimizer = original_optimizer

  def get_default_quantization_config(
      self, representative_data: object_detector_dataloader.DataLoader
  ) -> configs.QuantizationConfig:
    """Gets the default quantization configuration."""
    # Sets `inference_output_type=None` since the output op is the custom NMS op
    # which can't be quantized so that the tflite model output should be float.
    # `supported_ops` contains both `TFLITE_BUILTINS_INT8` and `TFLITE_BUILTINS`
    # due to the same reason: custom NMS op should be included as the
    # `TFLITE_BUILTINS` op.
    config = configs.QuantizationConfig.for_int8(
        representative_data,
        inference_output_type=None,
        supported_ops=[
            tf.lite.OpsSet.TFLITE_BUILTINS_INT8, tf.lite.OpsSet.TFLITE_BUILTINS
        ])
    config.experimental_new_quantizer = True
    return config

  def export_tflite(
      self,
      model: tf.keras.Model,
      tflite_filepath: str,
      quantization_config: Optional[configs.QuantizationConfig] = None) -> None:
    """Converts the retrained model to tflite format and saves it.

    The exported TFLite model has the following inputs & outputs:
    One input:
      image: a float32 tensor of shape[1, height, width, 3] containing the
        normalized input image. `self.config.image_size` is [height, width].

    Four Outputs:
      detection_boxes: a float32 tensor of shape [1, num_boxes, 4] with box
        locations.
      detection_classes: a float32 tensor of shape [1, num_boxes] with class
        indices.
      detection_scores: a float32 tensor of shape [1, num_boxes] with class
        scores.
      num_boxes: a float32 tensor of size 1 containing the number of detected
        boxes.

    Args:
      model: The EfficientDetNet model used for training which doesn't have pre
        and post processing.
      tflite_filepath: File path to save tflite model.
      quantization_config: Configuration for post-training quantization.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
      self.export_saved_model(
          model, temp_dir, batch_size=1, pre_mode=None, post_mode='tflite')
      converter = tf.lite.TFLiteConverter.from_saved_model(temp_dir)

      if quantization_config:
        converter = quantization_config.get_converter_with_quantization(
            converter, model_spec=self)

        # TFLITE_BUILTINS is needed for TFLite's custom NMS op for integer only
        # quantization.
        if tf.lite.OpsSet.TFLITE_BUILTINS not in converter.target_spec.supported_ops:
          converter.target_spec.supported_ops += [
              tf.lite.OpsSet.TFLITE_BUILTINS
          ]

      tflite_model = converter.convert()

      with tf.io.gfile.GFile(tflite_filepath, 'wb') as f:
        f.write(tflite_model)


efficientdet_lite0_spec = functools.partial(
    EfficientDetModelSpec,
    model_name='efficientdet-lite0',
    uri='https://tfhub.dev/tensorflow/efficientdet/lite0/feature-vector/1',
)
efficientdet_lite0_spec.__doc__ = util.wrap_doc(
    EfficientDetModelSpec,
    'Creates EfficientDet-Lite0 model spec. See also: `tflite_model_maker.object_detector.EfficientDetSpec`.'
)
mm_export('object_detector.EfficientDetLite0Spec').export_constant(
    __name__, 'efficientdet_lite0_spec')

efficientdet_lite1_spec = functools.partial(
    EfficientDetModelSpec,
    model_name='efficientdet-lite1',
    uri='https://tfhub.dev/tensorflow/efficientdet/lite1/feature-vector/1',
)
efficientdet_lite1_spec.__doc__ = util.wrap_doc(
    EfficientDetModelSpec,
    'Creates EfficientDet-Lite1 model spec. See also: `tflite_model_maker.object_detector.EfficientDetSpec`.'
)
mm_export('object_detector.EfficientDetLite1Spec').export_constant(
    __name__, 'efficientdet_lite1_spec')

efficientdet_lite2_spec = functools.partial(
    EfficientDetModelSpec,
    model_name='efficientdet-lite2',
    uri='https://tfhub.dev/tensorflow/efficientdet/lite2/feature-vector/1',
)
efficientdet_lite2_spec.__doc__ = util.wrap_doc(
    EfficientDetModelSpec,
    'Creates EfficientDet-Lite2 model spec. See also: `tflite_model_maker.object_detector.EfficientDetSpec`.'
)
mm_export('object_detector.EfficientDetLite2Spec').export_constant(
    __name__, 'efficientdet_lite2_spec')

efficientdet_lite3_spec = functools.partial(
    EfficientDetModelSpec,
    model_name='efficientdet-lite3',
    uri='https://tfhub.dev/tensorflow/efficientdet/lite3/feature-vector/1',
)
efficientdet_lite3_spec.__doc__ = util.wrap_doc(
    EfficientDetModelSpec,
    'Creates EfficientDet-Lite3 model spec. See also: `tflite_model_maker.object_detector.EfficientDetSpec`.'
)
mm_export('object_detector.EfficientDetLite3Spec').export_constant(
    __name__, 'efficientdet_lite3_spec')

efficientdet_lite4_spec = functools.partial(
    EfficientDetModelSpec,
    model_name='efficientdet-lite4',
    uri='https://tfhub.dev/tensorflow/efficientdet/lite4/feature-vector/2',
)
efficientdet_lite4_spec.__doc__ = util.wrap_doc(
    EfficientDetModelSpec,
    'Creates EfficientDet-Lite4 model spec. See also: `tflite_model_maker.object_detector.EfficientDetSpec`.'
)
mm_export('object_detector.EfficientDetLite4Spec').export_constant(
    __name__, 'efficientdet_lite4_spec')



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common Dataset used for tasks."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
from typing import Optional

import tensorflow as tf


def shard(ds, input_pipeline_context):
  # The dataset is always sharded by number of hosts.
  # num_input_pipelines is the number of hosts rather than number of cores.
  if (input_pipeline_context and
      input_pipeline_context.num_input_pipelines > 1):
    ds = ds.shard(input_pipeline_context.num_input_pipelines,
                  input_pipeline_context.input_pipeline_id)
  return ds


class DataLoader(object):
  """This class provides generic utilities for loading customized domain data that will be used later in model retraining.

  For different ML problems or tasks, such as image classification, text
  classification etc., a subclass is provided to handle task-specific data
  loading requirements.
  """

  def __init__(self, dataset, size=None):
    """Init function for class `DataLoader`.

    In most cases, one should use helper functions like `from_folder` to create
    an instance of this class.

    Args:
      dataset: A tf.data.Dataset object that contains a potentially large set of
        elements, where each element is a pair of (input_data, target). The
        `input_data` means the raw input data, like an image, a text etc., while
        the `target` means some ground truth of the raw input data, such as the
        classification label of the image etc.
      size: The size of the dataset. tf.data.Dataset donesn't support a function
        to get the length directly since it's lazy-loaded and may be infinite.
    """
    self._dataset = dataset
    self._size = size

  @property
  def size(self) -> Optional[int]:
    """Returns the size of the dataset.

    Note that this function may return None becuase the exact size of the
    dataset isn't a necessary parameter to create an instance of this class,
    and tf.data.Dataset donesn't support a function to get the length directly
    since it's lazy-loaded and may be infinite.
    In most cases, however, when an instance of this class is created by helper
    functions like 'from_folder', the size of the dataset will be preprocessed,
    and this function can return an int representing the size of the dataset.
    """
    return self._size

  def gen_dataset(self,
                  batch_size=1,
                  is_training=False,
                  shuffle=False,
                  input_pipeline_context=None,
                  preprocess=None,
                  drop_remainder=False):
    """Generate a shared and batched tf.data.Dataset for training/evaluation.

    Args:
      batch_size: A integer, the returned dataset will be batched by this size.
      is_training: A boolean, when True, the returned dataset will be optionally
        shuffled and repeated as an endless dataset.
      shuffle: A boolean, when True, the returned dataset will be shuffled to
        create randomness during model training.
      input_pipeline_context: A InputContext instance, used to shared dataset
        among multiple workers when distribution strategy is used.
      preprocess: A function taking three arguments in order, feature, label and
        boolean is_training.
      drop_remainder: boolean, whether the finaly batch drops remainder.

    Returns:
      A TF dataset ready to be consumed by Keras model.
    """
    ds = self._dataset
    ds = shard(ds, input_pipeline_context)

    if preprocess:
      preprocess = functools.partial(preprocess, is_training=is_training)
      ds = ds.map(preprocess, num_parallel_calls=tf.data.AUTOTUNE)

    if is_training:
      if shuffle:
        # Shuffle size should be bigger than the batch_size. Otherwise it's only
        # shuffling within the batch, which equals to not having shuffle.
        buffer_size = 3 * batch_size
        # But since we are doing shuffle before repeat, it doesn't make sense to
        # shuffle more than total available entries.
        # TODO(wangtz): Do we want to do shuffle before / after repeat?
        # Shuffle after repeat will give a more randomized dataset and mix the
        # epoch boundary: https://www.tensorflow.org/guide/data
        if self._size:
          buffer_size = min(self._size, buffer_size)
        ds = ds.shuffle(buffer_size=buffer_size)

    ds = ds.batch(batch_size, drop_remainder=drop_remainder)
    ds = ds.prefetch(tf.data.AUTOTUNE)
    # TODO(b/171449557): Consider converting ds to distributed ds here.
    return ds

  def __len__(self):
    if self._size is not None:
      return self._size
    else:
      return len(self._dataset)

  def split(self, fraction):
    """Splits dataset into two sub-datasets with the given fraction.

    Primarily used for splitting the data set into training and testing sets.

    Args:
      fraction: float, demonstrates the fraction of the first returned
        subdataset in the original data.

    Returns:
      The splitted two sub datasets.
    """
    return self._split(fraction)

  def _split(self, fraction, *args):
    """Actual implementation for `split` method and returns sub-class instances.

    Child DataLoader, if requires additional constructor arguments, should
      implement their own `split` method by calling `_split` with all arguments
      to the constructor.

    Args:
      fraction: float, demonstrates the fraction of the first returned
        subdataset in the original data.
      *args: additional arguments passed to the sub-class constructor.

    Returns:
      The splitted two sub datasets.
    """
    assert (fraction > 0 and fraction < 1)

    ds = self._dataset

    train_size = int(self._size * fraction)
    trainset = self.__class__(ds.take(train_size), train_size, *args)

    test_size = self._size - train_size
    testset = self.__class__(ds.skip(train_size), test_size, *args)

    return trainset, testset


class ClassificationDataLoader(DataLoader):
  """DataLoader for classification models."""

  def __init__(self, dataset, size, index_to_label):
    super(ClassificationDataLoader, self).__init__(dataset, size)
    self.index_to_label = index_to_label

  @property
  def num_classes(self):
    return len(self.index_to_label)

  def split(self, fraction):
    """Splits dataset into two sub-datasets with the given fraction.

    Primarily used for splitting the data set into training and testing sets.

    Args:
      fraction: float, demonstrates the fraction of the first returned
        subdataset in the original data.

    Returns:
      The splitted two sub datasets.
    """
    return self._split(fraction, self.index_to_label)



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Eval libraries. Used for TFLite model without post-processing."""
from absl import app
from absl import flags
from absl import logging
import numpy as np
import tensorflow as tf

from tensorflow_examples.lite.model_maker.third_party.efficientdet import coco_metric
from tensorflow_examples.lite.model_maker.third_party.efficientdet import dataloader
from tensorflow_examples.lite.model_maker.third_party.efficientdet import hparams_config
from tensorflow_examples.lite.model_maker.third_party.efficientdet import utils

from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import anchors
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import label_util
from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import postprocess

FLAGS = flags.FLAGS

DEFAULT_SCALE, DEFAULT_ZERO_POINT = 0, 0


def define_flags():
  """Define the flags."""
  flags.DEFINE_integer('eval_samples', None, 'Number of eval samples.')
  flags.DEFINE_string('val_file_pattern', None,
                      'Glob for eval tfrecords, e.g. coco/val-*.tfrecord.')
  flags.DEFINE_string('val_json_file', None,
                      'Groudtruth, e.g. annotations/instances_val2017.json.')
  flags.DEFINE_string('model_name', 'efficientdet-d0', 'Model name to use.')
  flags.DEFINE_string('tflite_path', None, 'Path to TFLite model.')
  flags.DEFINE_bool(
      'only_network', False,
      'TFLite model only contains EfficientDetNet without post-processing NMS op.'
  )
  flags.DEFINE_bool('pre_class_nms', False, 'Use pre_class_nms for evaluation.')
  flags.DEFINE_string('hparams', '', 'Comma separated k=v pairs or a yaml file')
  flags.mark_flag_as_required('val_file_pattern')
  flags.mark_flag_as_required('tflite_path')


class LiteRunner(object):
  """Runs inference with TF Lite model."""

  def __init__(self, tflite_model_path, only_network=False):
    """Initializes Lite runner with tflite model file.

    Args:
      tflite_model_path: Path to TFLite model file.
      only_network: boolean, If True, TFLite model only contains EfficientDetNet
        without post-processing NMS op. If False, TFLite model contains custom
        NMS op.
    """
    self.interpreter = tf.lite.Interpreter(tflite_model_path)
    self.interpreter.allocate_tensors()
    # Get input and output tensors.
    self.input_details = self.interpreter.get_input_details()
    self.output_details = self.interpreter.get_output_details()
    self.only_network = only_network
    if self.only_network:
      # TFLite with only_network=True changed the order of output. Sort it
      # according to tensor name.
      # TFLite with only_network=False  doesn't change the order since the
      # latest op is cutom NMS op.
      self.output_details = sorted(self.output_details, key=lambda a: a['name'])

  def run(self, image):
    """Runs inference with Lite model."""
    interpreter = self.interpreter
    signature_fn = interpreter.get_signature_runner()
    input_details = self.input_details
    output_details = self.output_details

    input_detail = input_details[0]
    if input_detail['quantization'] != (DEFAULT_SCALE, DEFAULT_ZERO_POINT):
      scale, zero_point = input_detail['quantization']
      image = image / scale + zero_point
      image = np.array(image, dtype=input_detail['dtype'])

    output = signature_fn(images=image)

    def get_output(idx):
      output_detail = output_details[idx]
      output_tensor = output[f'output_{idx}']
      if output_detail['quantization'] != (DEFAULT_SCALE, DEFAULT_ZERO_POINT):
        # Dequantize the output
        scale, zero_point = output_detail['quantization']
        output_tensor = output_tensor.astype(np.float32)
        output_tensor = (output_tensor - zero_point) * scale
      return output_tensor

    output_size = len(output_details)
    if not self.only_network:
      # TFLite model with post-processing.
      # Four Outputs:
      #   num_boxes: a float32 tensor of size 1 containing the number of
      #     detected boxes
      #   detection_scores: a float32 tensor of shape [1, num_boxes]
      #     with class scores
      #   detection_classes: a float32 tensor of shape [1, num_boxes]
      #     with class indices
      #   detection_boxes: a float32 tensor of shape [1, num_boxes, 4] with box
      #     locations
      return [get_output(i) for i in range(output_size)]
    else:
      # TFLite model only contains network without post-processing.
      num_boxes = int(output_size / 2)
      cls_outputs, box_outputs = [], []
      for i in range(num_boxes):
        cls_outputs.append(get_output(i))
        box_outputs.append(get_output(i + num_boxes))
      return cls_outputs, box_outputs


def main(_):
  config = hparams_config.get_efficientdet_config(FLAGS.model_name)
  config.override(FLAGS.hparams)
  config.val_json_file = FLAGS.val_json_file
  config.nms_configs.max_nms_inputs = anchors.MAX_DETECTION_POINTS
  config.drop_remainder = False  # eval all examples w/o drop.
  config.image_size = utils.parse_image_size(config['image_size'])

  # Evaluator for AP calculation.
  label_map = label_util.get_label_map(config.label_map)
  evaluator = coco_metric.EvaluationMetric(
      filename=config.val_json_file, label_map=label_map)

  # dataset
  batch_size = 1
  ds = dataloader.InputReader(
      FLAGS.val_file_pattern,
      is_training=False,
      max_instances_per_image=config.max_instances_per_image)(
          config, batch_size=batch_size)
  eval_samples = FLAGS.eval_samples
  if eval_samples:
    ds = ds.take((eval_samples + batch_size - 1) // batch_size)

  # Network
  lite_runner = LiteRunner(FLAGS.tflite_path, FLAGS.only_network)
  eval_samples = FLAGS.eval_samples or 5000
  pbar = tf.keras.utils.Progbar((eval_samples + batch_size - 1) // batch_size)
  for i, (images, labels) in enumerate(ds):
    if not FLAGS.only_network:
      nms_boxes_bs, nms_classes_bs, nms_scores_bs, _ = lite_runner.run(images)
      nms_classes_bs += postprocess.CLASS_OFFSET

      height, width = utils.parse_image_size(config.image_size)
      normalize_factor = tf.constant([height, width, height, width],
                                     dtype=tf.float32)
      nms_boxes_bs *= normalize_factor
      if labels['image_scales'] is not None:
        scales = tf.expand_dims(tf.expand_dims(labels['image_scales'], -1), -1)
        nms_boxes_bs = nms_boxes_bs * tf.cast(scales, nms_boxes_bs.dtype)
      detections = postprocess.generate_detections_from_nms_output(
          nms_boxes_bs, nms_classes_bs, nms_scores_bs, labels['source_ids'])
    else:
      cls_outputs, box_outputs = lite_runner.run(images)
      detections = postprocess.generate_detections(
          config,
          cls_outputs,
          box_outputs,
          labels['image_scales'],
          labels['source_ids'],
          pre_class_nms=FLAGS.pre_class_nms)

    detections = postprocess.transform_detections(detections)
    evaluator.update_state(labels['groundtruth_data'].numpy(),
                           detections.numpy())
    pbar.update(i)

  # compute the final eval results.
  metrics = evaluator.result()
  metric_dict = {}
  for i, name in enumerate(evaluator.metric_names):
    metric_dict[name] = metrics[i]

  if label_map:
    for i, cid in enumerate(sorted(label_map.keys())):
      name = 'AP_/%s' % label_map[cid]
      metric_dict[name] = metrics[i + len(evaluator.metric_names)]
  print(FLAGS.model_name, metric_dict)


if __name__ == '__main__':
  define_flags()
  logging.set_verbosity(logging.WARNING)
  app.run(main)



#   Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Context encoder layer for on-device personalized recommendation model."""
import math
from typing import Dict

import tensorflow as tf

from tensorflow_examples.lite.model_maker.third_party.recommendation.ml.configs import input_config_pb2
from tensorflow_examples.lite.model_maker.third_party.recommendation.ml.configs import model_config as model_config_class


def safe_div(x, y):
  return tf.where(tf.not_equal(y, 0), tf.divide(x, y), tf.zeros_like(x))


class ContextEncoder(tf.keras.layers.Layer):
  """Layer to encode context sequence.

  This encoder layer supports three types: 1) bow: bag of words style averaging
  sequence embeddings. 2) cnn: use convolutional neural network to encode
  sequence. 3) rnn: use recurrent neural network to encode sequence.

  This encoder should be initialized with a predefined embedding layer, encoder
  type and necessary parameters corresponding to the encoder type.
  """

  def __init__(self,
               input_config: input_config_pb2.InputConfig,
               model_config: model_config_class.ModelConfig):
    """Initialize ContextEncoder layer.

    Initialize ContextEncoder layer according to input config and model config,
    setting up feature group encoders and hidden layers.

    Args:
      input_config: The input config (input_config_pb2.InputConfig).
      model_config: The model config (model_config_class.ModelConfig).
    """
    super(ContextEncoder, self).__init__()
    self._input_config = input_config
    self._model_config = model_config
    self._final_embedding_dim = self._input_config.label_feature.embedding_dim
    self._feature_group_encoders = [
        FeatureGroupEncoder(feature_group, self._model_config,
                            self._final_embedding_dim)
        for feature_group in input_config.global_feature_groups
    ]
    self._feature_group_encoders.extend([
        FeatureGroupEncoder(feature_group, self._model_config,
                            self._final_embedding_dim)
        for feature_group in input_config.activity_feature_groups
    ])
    self._final_embedding_dim = self._input_config.label_feature.embedding_dim
    # Set up hidden layers.
    self._hidden_layers = []
    for i, layer_dim in enumerate(self._model_config.hidden_layer_dims):
      self._hidden_layers.append(
          tf.keras.layers.Dense(
              units=layer_dim,
              name='hidden_layer{}'.format(i),
              activation=tf.nn.relu))
    # Append final top layer, dimension of which should equal final embedding
    # dimension.
    self._hidden_layers.append(
        tf.keras.layers.Dense(
            units=self._final_embedding_dim,
            name='proj_layer',
            activation=tf.nn.relu))

  def call(self, input_context: Dict[str, tf.Tensor]) -> tf.Tensor:
    """Call function of the context encoder layer.

    Takes in input context tensor dictionary, generates feature group
    embeddings, concatenate them up and pass them through top hidden layers
    to get final context embedding.

    Args:
      input_context: A dictionary mapping input context feature name to tenors.

    Returns:
      Context encoding.
    """
    input_context = {
        k: tf.expand_dims(v, 0) if v.shape.ndims == 1 else v
        for k, v in input_context.items()
    }
    feature_group_embeddings = [
        feature_group_encoder(input_context)
        for feature_group_encoder in self._feature_group_encoders
    ]
    context_embedding = tf.concat(feature_group_embeddings, -1)
    for hidden_layer in self._hidden_layers:
      context_embedding = hidden_layer(context_embedding)
    return context_embedding


class FeatureGroupEncoder(tf.keras.layers.Layer):
  """Layer to generate encoding for the feature group.

  Layer to generate encoding for the group of features in the feature
  group with encoder type (BOW/CNN/LSTM) specified in the feature group config.
  Embeddings of INT or STRING type features are concatenated first, and FLOAT
  type feature values are appended after. Embedding vector is properly masked.
  """

  def __init__(self,
               feature_group: input_config_pb2.FeatureGroup,
               model_config: model_config_class.ModelConfig,
               final_embedding_dim: int):
    """Initialize FeatureGroupEncoder layer.

    Initialize ContextEncoder layer according to feature group.

    Args:
      feature_group: The feature group that needs to encode.
      model_config: The model config (model_config_class.ModelConfig).
      final_embedding_dim: Dimension of final context embedding.
    """
    super(FeatureGroupEncoder, self).__init__()
    self._feature_group = feature_group
    self._model_config = model_config
    self._final_embedding_dim = final_embedding_dim
    # Prepare embedding layers.
    self._embedding_layer_dict = {}
    self._nonembedding_feature_names = []
    for feature in self._feature_group.features:
      if feature.feature_type in [
          input_config_pb2.FeatureType.INT, input_config_pb2.FeatureType.STRING
      ]:
        assert feature.HasField('vocab_size')
        assert feature.HasField('embedding_dim')
        embedding_layer = tf.keras.layers.Embedding(
            feature.vocab_size,
            feature.embedding_dim,
            embeddings_initializer=tf.keras.initializers.truncated_normal(
                mean=0.0, stddev=1.0 / math.sqrt(feature.embedding_dim)),
            mask_zero=True,
            name=feature.feature_name+'embedding_layer')
        self._embedding_layer_dict[feature.feature_name] = embedding_layer
      else:
        self._nonembedding_feature_names.append(feature.feature_name)
  # Prepare CNN layers.
    self._conv1d_layers = []
    if self._feature_group.encoder_type == input_config_pb2.EncoderType.CNN:
      assert self._model_config.conv_num_filter_ratios
      assert self._model_config.conv_kernel_size
      for ratio in self._model_config.conv_num_filter_ratios:
        self._conv1d_layers.append(
            tf.keras.layers.Conv1D(
                filters=self._final_embedding_dim * ratio,
                kernel_size=self._model_config.conv_kernel_size,
                strides=self._model_config.conv_kernel_size,
                padding='same',
                activation='relu'))
    # Prepare LSTM layer.
    elif self._feature_group.encoder_type == input_config_pb2.EncoderType.LSTM:
      assert self._model_config.lstm_num_units
      self._lstm_layer = tf.keras.layers.LSTM(
          self._model_config.lstm_num_units)

  def call(self, input_context: Dict[str, tf.Tensor]) -> tf.Tensor:
    """Call function of the feature group encoder layer.

    Takes in input context tensor dictionary, generates embedding for features
    in the group, concatenate them up and encode them according to the
    specified encoder type.

    Feature embedding layers will have mask_zero enabled, and generated feature
    embeddings will be masked if any of the feature value is 0. In most
    cases, the mask of all features in the same group should be the same. This
    function treats the input config generically.

    Args:
      input_context: A dictionary mapping input context feature name to tenors.

    Returns:
      Feature group encoding.
    """
    embedding_feature_shape = []
    embeddings = []
    masks = []
    for feature_name, embedding_layer in sorted(
        self._embedding_layer_dict.items()):
      input_feature = input_context[feature_name]
      if isinstance(input_feature, tf.SparseTensor):
        input_feature = tf.sparse.to_dense(input_feature)
      if not embedding_feature_shape:
        embedding_feature_shape = input_feature.shape.as_list()
      assert input_feature.shape.as_list() == embedding_feature_shape

      embeddings.append(embedding_layer(input_feature))
      masks.append(embedding_layer.compute_mask(input_feature))
    # Collect nonembedding feature values.
    for feature_name in self._nonembedding_feature_names:
      input_feature = input_context[feature_name]
      if isinstance(input_feature, tf.SparseTensor):
        input_feature = tf.sparse.to_dense(input_feature)
      embeddings.append(tf.expand_dims(input_feature, -1))
    embedding = tf.concat(embeddings, -1)

    # Set mask_shape with batch dim = 1. Compatible to unknown batch size.
    mask_shape = [1] + embedding_feature_shape[1:]
    mask = tf.ones(shape=mask_shape)
    for m in masks:
      mask = mask * tf.cast(m, 'float32')
    mask = tf.expand_dims(mask, -1)
    embedding = embedding * mask

    if self._feature_group.encoder_type == input_config_pb2.EncoderType.BOW:
      weighted_summed_context_embedding = tf.keras.backend.sum(
          embedding, axis=1)
      total_weights = tf.ones_like(
          weighted_summed_context_embedding) * tf.keras.backend.sum(
              mask, axis=1)
      embedding = safe_div(weighted_summed_context_embedding, total_weights)
    elif self._feature_group.encoder_type == input_config_pb2.EncoderType.CNN:
      for conv1d_layer in self._conv1d_layers:
        embedding = conv1d_layer(embedding)
      embedding = tf.math.reduce_max(embedding, axis=1)
    elif self._feature_group.encoder_type == input_config_pb2.EncoderType.LSTM:
      embedding = self._lstm_layer(embedding)
    else:
      raise ValueError('Unsupported encoder type %s.' %
                       self._feature_group.encoder_type)

    return embedding
